@@ -1,18 +1,20 @@
-import random
 import warnings
 from typing import Callable, Optional, Dict, Tuple, Union
 import numpy as np
 import pandas as pd
 from ordered_set import OrderedSet
 from typeguard import typechecked
+from tqdm import tqdm
 
 from msapy import utils as ut
+from msapy.checks import _check_valid_elements, _check_valid_n_permutations, _check_valid_permutation_space, _get_contribution_type
 
 
 @typechecked
 def make_permutation_space(*,
                            elements: list,
                            n_permutations: int,
+                           rng: Optional[np.random.Generator] = None,
                            random_seed: Optional[int] = None) -> list:
     """
     Generates a list of tuples containing n_permutations of the given elements.
@@ -42,25 +44,23 @@ def make_permutation_space(*,
             Number of permutations, Didn't check it systematically yet but just based on random explorations I'd say
             something around 1_000 is enough.
 
-        random_seed (Optional[int]):
-            sets the random seed of the sampling process. Default is None so if nothing is given every call results in
-            a different orderings.
+        rng (Optional[np.random.Generator]): Numpy random generator object used for reproducable results. Default is None.
 
-    Returns (list[tuple]):
-        Permutation space as a list of lists with shape (n_permutations, len(elements))
+        random_seed (Optional[int]):
+            sets the random seed of the sampling process. Only used when `rng` is None. Default is None.
+
+    Returns:
+        (list[tuple]): Permutation space as a list of lists with shape (n_permutations, len(elements))
     """
 
     # ------------------------------#
-    if n_permutations <= 0:
-        raise ValueError("Specified number of permutations doesn't make sense because it's either zero or smaller.")
-    elif 1 < n_permutations < 100:
-        warnings.warn("Specified number of permutations is kinda small so the results might not be as accurate.",
-                      stacklevel=2)
+    _check_valid_elements(elements)
+    _check_valid_n_permutations(n_permutations)
     # ------------------------------#
-    if random_seed:
-        random.seed(random_seed)
+    if not rng:
+        rng = np.random.default_rng(random_seed) if random_seed else np.random.default_rng()
 
-    permutation_space = [tuple(random.sample(elements, len(elements))) for _ in range(n_permutations)]
+    permutation_space = [tuple(rng.permutation(elements)) for _ in range(n_permutations)]
     return permutation_space
 
 
@@ -89,12 +89,14 @@ def make_combination_space(*, permutation_space: list) -> OrderedSet:
         permutation_space (list):
             A list of players to be shuffled n times.
 
-    Returns (OrderedSet):
-        Combination space as an OrderedSet of frozensets.
+    Returns:
+        (OrderedSet): Combination space as an OrderedSet of frozensets.
     """
 
+    _check_valid_permutation_space(permutation_space)
+
     combination_space = OrderedSet()
-    for permutation in ut.generatorize(to_iterate=permutation_space):
+    for permutation in permutation_space:
 
         for index, _ in enumerate(permutation):  # we really don't care about the element itself here
 
@@ -124,9 +126,10 @@ def make_complement_space(*,
         elements (list):
             list of players.
 
-    returns (OrderedSet):
-        complements to be passed for lesioning.
+    Returns:
+        (OrderedSet): complements to be passed for lesioning.
     """
+    _check_valid_elements(elements)
 
     elements = frozenset(elements)
     diff = combination_space.items[len(elements)] ^ elements
@@ -138,7 +141,7 @@ def make_complement_space(*,
     # ------------------------------#
 
     complement_space = OrderedSet()
-    for combination in ut.generatorize(to_iterate=combination_space):
+    for combination in combination_space:
         complement_space.add(tuple(elements.difference(combination)))
     return complement_space
 
@@ -208,8 +211,8 @@ def take_contributions(*,
         objective_function_params (Optional[Dict]):
             Kwargs for the objective_function.
 
-    Returns (Dict):
-        A dictionary of combinations:results
+    Returns:
+        (Dict): A dictionary of combinations:results
     """
 
     elements = frozenset(elements)
@@ -225,7 +228,7 @@ def take_contributions(*,
                       stacklevel=2)
     # ------------------------------#
 
-    for combination in ut.generatorize(to_iterate=combination_space):
+    for combination in tqdm(combination_space):
         complement = tuple(elements.difference(combination))  # lesion everything but the target coalition
         result = objective_function(complement, **objective_function_params)
 
@@ -236,7 +239,7 @@ def take_contributions(*,
 
 @typechecked
 def make_shapley_values(*,
-                        contributions: Union[Dict, Tuple[Dict, Dict]],
+                        contributions: Dict,
                         permutation_space: list) -> pd.DataFrame:
     """
     Calculates Shapley values based on the filled contribution_space.
@@ -256,37 +259,59 @@ def make_shapley_values(*,
         permutation_space (list):
             Should be the same passed to make_combination_space.
 
-    Returns (pd.DataFrame):
-        Shapley table, columns will be elements and indices will be samples (permutations).
+    Returns:
+        pd.DataFrame: Shapley table or a dict of Shapely tables, columns will be 
+        elements and indices will be samples (permutations). 
+        It will be a Multi-Index DataFrame if the contributions are a dict
+        i.e. the objective function returns multiple score functions (eg. accuracy, f1_score, etc.)
+        It will be a Multi-Index DataFrame if the contributions are a timeseries.
+        The index at `level=1` will be the timestamps
     """
+    _check_valid_permutation_space(permutation_space)
+    arbitrary_contrib, multi_scores, is_timeseries = _get_contribution_type(contributions)
+
+    if multi_scores:
+        scores = list(arbitrary_contrib.keys())
+        contributions = {k: np.array(list(v.values())) for k, v in contributions.items()}
+
     shapley_table = {}
-    for permutation in ut.generatorize(to_iterate=permutation_space):
+    for permutation in permutation_space:
         isolated_contributions = []  # got to be a better way!
 
         # if the set is small it's possible that the permutation space exhausts the combination space so:
         if permutation not in shapley_table:
-            for index, element in enumerate(ut.generatorize(to_iterate=permutation)):
+            for index, element in enumerate(permutation):
                 including = frozenset(permutation[:index + 1])
                 excluding = frozenset(permutation[:index])
-
                 isolated_contributions.append(contributions[including] - contributions[excluding])
-            shapley_table[permutation] = isolated_contributions
+            shapley_table[permutation] = np.array(isolated_contributions)
+
+    if not multi_scores:
+        shapley_values = pd.DataFrame([
+            dict(zip(permutations, shapleys)) for permutations, shapleys in shapley_table.items()])
+        return _process_timeseries_shapley(shapley_values) if is_timeseries else shapley_values
+
+    shapley_values = []
+    for i in range(len(arbitrary_contrib)):
+        shapley_values.append(pd.DataFrame([
+            dict(zip(permutations, shapleys[:, i])) for permutations, shapleys in shapley_table.items()]))
+
+    shapley_values = pd.concat(shapley_values, keys=scores) if multi_scores else shapley_values
 
-    shapley_values = pd.DataFrame([
-        dict(zip(permutations, shapleys)) for permutations, shapleys in shapley_table.items()])
     return shapley_values
 
 
 @typechecked
 def interface(*,
               n_permutations: int,
-              n_parallel_games: int = -1,
               elements: list,
               objective_function: Callable,
-              objective_function_params: Optional[Dict] = None,
+              objective_function_params: Dict = {},
               permutation_space: Optional[list] = None,
               multiprocessing_method: str = 'joblib',
+              rng: Optional[np.random.Generator] = None,
               random_seed: Optional[int] = None,
+              n_parallel_games: int = -1,
               ) -> Tuple[pd.DataFrame, Dict, Dict]:
     """
     A wrapper function to call other related functions internally and produces an easy-to-use pipeline.
@@ -295,11 +320,6 @@ def interface(*,
         n_permutations (int):
             Number of permutations (samples) per element.
 
-        n_parallel_games (int):
-            Number of parallel jobs (number of to-be-occupied cores),
-            -1 means all CPU cores and 1 means a serial process.
-            I suggest using 1 for debugging since things gets messy in parallel!
-
         elements (list):
             List of the players (elements). Can be strings (names), integers (indicies), and tuples.
 
@@ -326,7 +346,7 @@ def interface(*,
                     lesioned.remove_nodes_from(complements)
                     return nx.local_efficiency(lesioned)
 
-        objective_function_params (Optional[Dict]):
+        objective_function_params (Dict):
            Kwargs for the objective_function.
 
         permutation_space (Optional[list]):
@@ -351,24 +371,32 @@ def interface(*,
 
             That makes sense since I have 16 cores and 1000/16 is around 62.
 
+        rng (Optional[np.random.Generator]): Numpy random generator object used for reproducable results. Default is None.
+
         random_seed (Optional[int]):
-            sets the random seed of the sampling process. Default is None so if nothing is given every call results in
-            a different orderings.
+            sets the random seed of the sampling process. Only used when `rng` is None. Default is None.
 
-    Returns ([pd.DataFrame, Dict, Dict]):
-        shapley_table, contributions, lesion_effects
+        n_parallel_games (int):
+            Number of parallel jobs (number of to-be-occupied cores),
+            -1 means all CPU cores and 1 means a serial process.
+            I suggest using 1 for debugging since things get messy in parallel!
 
-        Note that contributions and lesion_effects are the same values, addressed differently. For example:
-        If from a set of ABCD removing AC ends with some value x, you can say the contribution of BD=x and the
-        effect of removing AC=x. So the same values are addressed differently in the two returned Dicts.
-        Of course, it makes more sense to compare the lesion effects with the intact system but who am I to judge.
+    Returns:
+        Tuple[pd.DataFrame, Dict, Dict]: shapley_table, contributions, lesion_effects
+
+    Note that contributions and lesion_effects are the same values, addressed differently. For example:
+    If from a set of ABCD removing AC ends with some value x, you can say the contribution of BD=x and the
+    effect of removing AC=x. So the same values are addressed differently in the two returned Dicts.
+    Of course, it makes more sense to compare the lesion effects with the intact system but who am I to judge.
     """
-    of_params = objective_function_params if objective_function_params else {}
+
+    if not rng:
+        rng = np.random.default_rng(random_seed) if random_seed else np.random.default_rng()
 
     if not permutation_space:
         permutation_space = make_permutation_space(elements=elements,
                                                    n_permutations=n_permutations,
-                                                   random_seed=random_seed)
+                                                   rng=rng)
     else:
         warnings.warn("A Permutation space is given so n_permutations will fall back to what's specified there.",
                       stacklevel=2)
@@ -382,7 +410,7 @@ def interface(*,
                                                            complement_space=complement_space,
                                                            combination_space=combination_space,
                                                            objective_function=objective_function,
-                                                           objective_function_params=of_params)
+                                                           objective_function_params=objective_function_params)
     else:
         contributions, lesion_effects = ut.parallelized_take_contributions(
             multiprocessing_method=multiprocessing_method,
@@ -390,16 +418,16 @@ def interface(*,
             complement_space=complement_space,
             combination_space=combination_space,
             objective_function=objective_function,
-            objective_function_params=of_params)
+            objective_function_params=objective_function_params)
 
-    shapley_values = make_shapley_values(contributions=contributions, permutation_space=permutation_space)
+    shapley_values = make_shapley_values(contributions=contributions, permutation_space=permutation_space)[elements]
     return shapley_values, contributions, lesion_effects
 
 
 @typechecked
-def estimate_causal_influences(network_connectome: np.ndarray,
+def estimate_causal_influences(elements: list,
                                objective_function: Callable,
-                               objective_function_params: Optional[Dict],
+                               objective_function_params: Optional[dict] = None,
                                multiprocessing_method: str = 'joblib',
                                n_cores: int = -1,
                                n_permutations: int = 1000,
@@ -414,23 +442,20 @@ def estimate_causal_influences(network_connectome: np.ndarray,
     the average activity of C.
 
     VERY IMPORTANT NOTES:
+
         1. The resulting causal contribution matrix does not necessarily reflect the connectome. In the example above
         there's no actual connection A -> C but there might be one in the causal contribution matrix since A is causally
         influencing C via B.
-
         2. Think twice (even three times) about your objective function. The same everything will result in different
         causal contribution matrices depending on what are you tracking and how accurate it's capturing the effect of
         lesions. Also don't forget the edge-cases. There will be weird behaviors in your system, for example, what it
         does if every node is perturbed?
-
         3. The metric you track is preferred to be non-negative and bounded (at least practically!)
-
         4. Obviously this will take N times longer than a normal MSA with N is the number of nodes. So make sure your
         process is as fast as it can be for example use Numba and stuff, but you don't need to implement any parallel
         processes since it's already implemented here. Going below 1000 permutations might be an option depending on
-        your specific case but based on experience, it's not a good idea.
-
-        and 5. Shapley values sum up (or will be close) to the value of the intact coalition. So for example if the
+        your specific case but based on experience, it's not a good idea 
+        5. Shapley values sum up (or will be close) to the value of the intact coalition. So for example if the
         mean activity of node C here is 50 then causal_contribution_matrix.sum(axis=0) = 50 or close to 50. If not it
         means:
             1. the number of permutations are not enough
@@ -439,9 +464,8 @@ def estimate_causal_influences(network_connectome: np.ndarray,
 
 
     Args:
-        network_connectome (np.ndarray):
-            Structural backbone of your network.
-            I probably can turn this into a list of elements instead so #TODO.
+        elements (list):
+            List of the players (elements). Can be strings (names), integers (indicies), and tuples.
 
         objective_function (Callable):
             The game (in-silico experiment). It should get the complement set and return one numeric value
@@ -449,7 +473,6 @@ def estimate_causal_influences(network_connectome: np.ndarray,
             This function is just calling it as: objective_function(complement, **objective_function_params)
 
             An example using networkx with some tips:
-            (you sometimes need to specify what should happen during edge-cases like an all-lesioned network)
 
             def lesion_me_senpai(complements, network, index):
                 # note "index", your function should be able to track the effects on the target and the keyword for
@@ -465,6 +488,9 @@ def estimate_causal_influences(network_connectome: np.ndarray,
                 activity = network.run(lesioned_network) # or really, whatever you want!
                 return float(activity[index].mean())
 
+            (you sometimes need to specify what should happen during edge-cases like an all-lesioned network)
+
+
         objective_function_params (Optional[Dict]):
             Kwargs for the objective_function. A dictionary pair of {'index': index} will be added to this during
             the process so your function can track the lesion effect.
@@ -489,50 +515,60 @@ def estimate_causal_influences(network_connectome: np.ndarray,
             a different orderings.
 
     Returns:
+        causal_influences (pd.DataFrame)
 
     """
-
-    # Initialize the stuff
-    nodes = [node for node,_ in enumerate(network_connectome)]
-    permutations = dict()
-    combinations = dict()
-    complements = dict()
-    contributions = dict()
-    lesion_effects = dict()
-    shapley_values = dict()
     objective_function_params = objective_function_params if objective_function_params else {}
 
+    # Initialize the stuff
+    shapley_values = []
     # Looping through the nodes
-    for index, _ in enumerate(network_connectome):
-        print(f"working on the node number {index} from {len(network_connectome)} nodes.")
+    for index, element in enumerate(elements):
+        print(f"working on node number {index} from {len(elements)} nodes.")
         objective_function_params['index'] = index
 
         # Takes the target out of the to_be_lesioned list
-        without_target = set(nodes).difference({index})
+        without_target = set(elements).difference({element})
 
-        # Generates N permutations and their respective combinations/complements of the network without the target.
-        permutations[index] = make_permutation_space(n_permutations=n_permutations,
-                                                     elements=list(without_target),
+        shapley_value, contributions, _ = interface(n_permutations=n_permutations,
+                                                     elements=list(
+                                                         without_target),
+                                                     objective_function=objective_function,
+                                                     objective_function_params=objective_function_params,
+                                                     n_parallel_games=n_cores,
+                                                     multiprocessing_method=multiprocessing_method,
                                                      random_seed=permutation_seed)
-        combinations[index] = make_combination_space(permutation_space=permutations[index])
-        complements[index] = make_complement_space(combination_space=combinations[index],
-                                                   elements=list(without_target))
 
-        # Plays the game for each lesion combination
-        contributions[index], lesion_effects[index] = ut.parallelized_take_contributions(
-            multiprocessing_method=multiprocessing_method,
-            n_cores=n_cores,
-            complement_space=complements[index],
-            combination_space=combinations[index],
-            objective_function=objective_function,
-            objective_function_params=objective_function_params)
+        _, multi_scores, is_timeseries = _get_contribution_type(
+            contributions)
 
-        # Calculates the good-old Shapley values for the source nodes
-        shapley_values[index] = make_shapley_values(contributions=contributions[index],
-                                                    permutation_space=permutations[index])
+        if multi_scores:
+            shapley_value = shapley_value.groupby(level=0).mean()
+        elif is_timeseries:
+            shapley_value = shapley_value.groupby(level=1).mean()
+        else:
+            shapley_value = shapley_value.mean()
+
+        shapley_values.append(shapley_value)
+
+    causal_influences = pd.concat(shapley_values, keys=elements) if (
+        multi_scores or is_timeseries) else pd.DataFrame(shapley_values, columns=elements)
+    return causal_influences.swaplevel().sort_index(level=0) if multi_scores else causal_influences
+
+
+@typechecked
+def _process_timeseries_shapley(shapley_values: pd.DataFrame) -> pd.DataFrame:
+    num_permutation, num_nodes = shapley_values.shape
+    data = np.stack(shapley_values.values.flatten())
+    num_timestamps = data.shape[-1]
+    data = data.reshape(num_permutation, num_nodes, -1)
+    data = data.transpose((0, 2, 1)).reshape((-1, num_nodes))
+
+    shapley_values = pd.DataFrame(data=data,
+                                  index=pd.MultiIndex.from_product(
+                                      [range(num_permutation), range(num_timestamps)], names=[None, "timestamp"]),
+                                  columns=shapley_values.columns
+                                  )
+
+    return shapley_values
 
-    # The value orders are not sorted so
-    for shapley_value in shapley_values:
-        shapley_values[shapley_value] = shapley_values[shapley_value].sort_index(axis=1).mean(axis=0)
-    causal_influences = pd.DataFrame.from_dict(shapley_values)
-    return causal_influences
